#include "types.h"
#include "param.h"
#include "memlayout.h"
#include "riscv.h"
#include "atomic/spinlock.h"
#include "kernel/proc.h"
#include "utils.h"
#include "defs.h"
#include "mm/vm.h"
#include "mm/mm.h"

#include "debug.h"
#include "fs/fcntl.h"
#include "fs/stat.h"
#include "fs/fs.h"
#include "fs/file.h"

#define max(a, b) ((a) > (b) ? (a) : (b))
#define min(a, b) ((a) < (b) ? (a) : (b))

struct cpu cpus[NUM_CORES];

struct proc proc[NPROC];

struct proc *initproc;

int nextpid = 1;
struct spinlock pid_lock;

extern void forkret(void);
static void freeproc(struct proc *p);

extern char trampoline[]; // trampoline.S

extern uint64_t sys_memuse(void);
void
freewalk(pagetable_t pagetable);

mm_struct_t *mm_init(mm_struct_t *mm, proc_t *tsk);
static void init_new_context(proc_t *tsk);
// helps ensure that wakeups of wait()ing
// parents are not lost. helps obey the
// memory model when using p->parent.
// must be acquired before any p->lock.
struct spinlock wait_lock;

// Allocate a page for each process's kernel stack.
// Map it high in memory, followed by an invalid
// guard page.
// void
// proc_mapstacks() {
//   struct proc *p;
  
//   for(p = proc; p < &proc[NPROC]; p++) {
//     char *pa = kalloc();
//     if(pa == 0)
//       panic("kalloc");
//     uint64 va = KSTACK((int) (p - proc));
//     kvmmap(va, (uint64)pa, PGSIZE, PTE_R | PTE_W, PGSPEC_NORMAL);
//   }
// }

// initialize the proc table at boot time.
void
procinit(void)
{
  struct proc *p;
  
  initlock(&pid_lock, "nextpid");
  initlock(&wait_lock, "wait_lock");
  for(p = proc; p < &proc[NPROC]; p++) {
      initlock(&p->lock, "proc");
      p->state = UNUSED;
      // p->kstack = KSTACK((int) (p - proc));
      for(int i = 0; i < NOFILE; i++) {
        p->ofile[i] = NULL;
      }
  }
  for(int i = 0; i < NUM_CORES; i++) {
    cpus[i].intena = 0;
    cpus[i].noff = 0;
    cpus[i].proc = NULL;
  }
}

// Return this CPU's cpu struct.
// Interrupts must be disabled.
struct cpu*
mycpu(void) {
  int id = cpuid();
  struct cpu *c = &cpus[id];
  return c;
}

// Return the current struct proc *, or zero if none.
struct proc*
myproc(void) {
  push_off();
  struct cpu *c = mycpu();
  struct proc *p = c->proc;
  pop_off();
  return p;
}

int
allocpid() {
  int pid;
  
  acquire(&pid_lock);
  pid = nextpid;
  nextpid = nextpid + 1;
  release(&pid_lock);

  return pid;
}

void
alloc_init_mm(struct proc *p){
  p->mm = kzalloc(sizeof(struct mm_struct));
  p->mm->mmap = NULL;
  p->mm->mm_rb.rb_node = NULL;
  p->mm->mmap_cache = NULL;
  p->mm->map_count = 0;
  initlock(&p->mm->mm_lock, "mm_lock");
  p->mm->free_area_cache = TASK_UNMAPPED_BASE;
  // p->mm->free_area_cache = 0;
}

// Look in the process table for an UNUSED proc.
// If found, initialize state required to run in the kernel,
// and return with p->lock held.
// If there are no free procs, or a memory allocation fails, return 0.
static struct proc*
allocproc(void)
{
  struct proc *p;

  for(p = proc; p < &proc[NPROC]; p++) {
    acquire(&p->lock);
    if(p->state == UNUSED) {
      goto found;
    } else {
      release(&p->lock);
    }
  }
  return 0;

found:
  p->pid = allocpid();
  p->state = USED;
  p->nfd = NOFILE;
  p->ext_ofile = NULL;

  // 申请Trapframe
  if((p->trapframe = (struct trapframe *)kalloc()) == 0){
    // freeproc(p);
    // release(&p->lock);
    // return 0;
    ER();
  }
  // 申请内核栈
  if((p->kstack = (uint64_t)kalloc()) == 0){
    // freeproc(p);
    // release(&p->lock);
    // return 0;
    ER();
  }

  // 获取空页表（已经映射了内核地址空间，trapfram以及内核栈）
  // p->pagetable = proc_pagetable(p);
  // if(p->pagetable == 0){
    // freeproc(p);
    // release(&p->lock);
    // return 0;
  // }

  // Set up new context to start executing at forkret,
  // which returns to user space.
  // memset(&p->context, 0, sizeof(p->context));
  // p->context.ra = (uint64)forkret;
  // p->context.sp = p->kstack + PGSIZE;

  // p->cur_mmap_sz = MMAP_BASE;
  // alloc_init_mm(p);

  return p;
}

// free a proc structure and the data hanging from it,
// including user pages.
// p->lock must be held.
static void
freeproc(struct proc *p)
{
  if(p->trapframe)
    kfree((void*)p->trapframe);
  if(p->kstack)
    kfree((void*)p->kstack);
  if(p->ext_ofile)
    kfree((void *)p->ext_ofile);
  p->trapframe = 0;
  p->kstack = 0;

  // free_vma(p);

  // for(;;);

  if(p->mm){
    /* 释放用户空间 */
    exit_mm(p);
    /*exit_mm 造成 user_trap*/
    // ER();
    /* 释放内核空间 */
    // proc_freepagetable(p->mm->pagetable);
    // p->mm = 0;
  } 
  // p->sz = 0;
  p->pid = 0;
  p->parent = 0;
  p->name[0] = 0;
  p->chan = 0;
  p->killed = 0;
  p->xstate = 0;
  p->state = UNUSED;
}

// Create a user page table for a given process,
// with no user memory, but with trampoline pages.
pagetable_t
proc_pagetable(mm_struct_t *mm, proc_t *tsk)
{
  pagetable_t pagetable;

  // An empty page table with kernel page table copy.
  pagetable = uvmcreate();
  if(pagetable == 0)
    return 0;
  // map the trampoline code (for system call return)
  // at the highest user virtual address.
  // only the supervisor uses it, on the way
  // to/from user space, so not PTE_U.
  // if(mappages(pagetable, TRAMPOLINE, PGSIZE,
  //             (uint64)trampoline, PTE_R | PTE_X) < 0){
  //   uvmfree(pagetable, 0);
  //   return 0;
  // }

  // map the trapframe just below TRAMPOLINE, for trampoline.S.
  if(mappages(pagetable, TRAPFRAME, PGSIZE,
              (uint64)(tsk->trapframe), PTE_R | PTE_W) < 0){
    erasekvm(pagetable);
    // uvmfree(pagetable, 0, MMAP_BASE);
    ER();
    return 0;
  }

  // if(mappages(pagetable, KSTACK, PGSIZE,
  //             (uint64)(tsk->kstack), PTE_R | PTE_W) < 0){
  //   erasekvm(pagetable);
  //   uvmunmap(pagetable, TRAPFRAME, 1, 0);
  //   // uvmfree(pagetable, 0, MMAP_BASE);
  //   ER();
  //   return 0;
  // }

  return pagetable;
}

// Free a process's page table, and free the
// physical memory it refers to.
void
proc_freepagetable(pagetable_t pagetable)
{
  uvmunmap(pagetable, TRAPFRAME, 1, 0);
  // uvmunmap(pagetable, KSTACK, 1, 0);
  erasekvm(pagetable);
  // uvmfree(pagetable, sz);
}

// a user program that calls exec("/init")
// od -t xC initcode
uchar initcode[] = {
  #include "generated/initcode_bin.h"
};

#include "fs/file.h"
static void init_std(proc_t *p) {
  struct file *stdin = filealloc();
  struct file *stdout = filealloc();
  struct file *stderr = filealloc();
  
  p->ofile[0] = stdin;
  p->ofile[1] = stdout;
  p->ofile[2] = stderr;

  stdin->type = FD_DEVICE;
  stdout->type = FD_DEVICE;
  stderr->type = FD_DEVICE;
  stdin->readable = 1;
  stdin->writable = 0;
  stdout->readable = 0;
  stdout->writable = 1;
  stderr->readable = 0;
  stderr->writable = 1;
  stdin->dev = &devs[CONSOLE];
  stdout->dev = &devs[CONSOLE];
  stderr->dev = &devs[CONSOLE];
}

// Set up first user process.
void
userinit(void)
{
  struct proc *p;

  p = allocproc();
  initproc = p;

  p->mm = kzalloc(sizeof(mm_struct_t));
  mm_init(p->mm, p);
  init_new_context(p);

  mycpu()->proc = p;
  // allocate one user page and copy init's instructions
  // and data into it.
  uvminit(p->mm->pagetable, initcode, sizeof(initcode));
  do_mmap(NULL, USER_START, PGSIZE, PROT_EXEC|PROT_READ|PROT_WRITE, MAP_EXECUTABLE, 0, LOAD);
  // p->sz = PGSIZE;

  // prepare for the very first "return" from kernel to user.
  p->trapframe->epc = USER_START;      // user program counter

  /* 栈是否过小，要不奥为其创建一个vma？ */
  p->trapframe->sp = PGSIZE;  // user stack pointer

  safestrcpy(p->name, "initcode", sizeof(p->name));
    // p->cwd = namei("/");

  init_std(p);

  p->state = RUNNABLE;

  release(&p->lock);
}


// int
// growproc(int n)
// {
//   uint sz;
//   struct proc *p = myproc();

//   sz = p->sz;
//   if(n > 0){
//     if((sz = uvmalloc(p->pagetable, sz, sz + n)) == 0) {
//       return -1;
//     }
//   } else if(n < 0){
//     sz = uvmdealloc(p->pagetable, sz, sz + n);
//   }
//   p->sz = sz;
//   return 0;
// }

/**
 * @brief copy the vm_area_struct of oldmm to mm,  also copy_page_range
 * 复制vma和页表
 * @param mm 
 * @param oldmm 
 * @return int 
 */
static int 
dup_mmap(mm_struct_t *mm, mm_struct_t *oldmm){

  vm_area_struct_t *mpnt, *tmp, **pprev;
  struct rb_node **rb_link, *rb_parent;
  // int retval;

  acquire(&oldmm->mm_lock);
  mm->mmap = NULL;
  mm->mmap_cache = NULL;
  mm->free_area_cache = oldmm->free_area_cache;
  mm->map_count = 0;
  mm->rss = oldmm->rss;
  mm->mm_rb.rb_node = NULL;

  rb_link = &mm->mm_rb.rb_node;
	rb_parent = NULL;

  pprev = &mm->mmap;

  /* 遍历oldmm中所有的vma */
  for(mpnt = oldmm->mmap; mpnt != NULL; mpnt = mpnt->vm_next){
    if(mpnt == NULL) 
      break;
    
    struct file *f;

    tmp = kzalloc(sizeof(vm_area_struct_t));
    if(!tmp)
      panic("do mmap1!");

    *tmp = *mpnt;
    tmp->vm_mm = mm;
    tmp->vm_next = NULL;
    f = tmp->vm_file;
    if(f){
      filedup(f);
    }

    /* 新分配的vma插入到链表 */
    acquire(&mm->page_table_lock);
    *pprev = tmp;
    pprev = &tmp->vm_next;

    /* 新分配的vma插入到红黑树 */
    __vma_link_rb(mm, tmp, rb_link, rb_parent);
		rb_link = &tmp->vm_rb.rb_right;
		rb_parent = &tmp->vm_rb;

    mm->map_count++;
    /* 复制页表到新的mm中 */
    uvmcopy(oldmm->pagetable, mm->pagetable, tmp->vm_start, tmp->vm_end);
    release(&mm->page_table_lock);
  } 

  release(&oldmm->mm_lock);
  return 0;
}

/**
 * @brief 初始化复制的mm结构体，并且分配新页表
 * 
 */
mm_struct_t *mm_init(mm_struct_t *mm, proc_t *tsk){
  initlock(&mm->page_table_lock, "");
  initlock(&mm->mm_lock, "");
  mm->free_area_cache = TASK_UNMAPPED_BASE;
  
  // 申请Trapframe
  // if((mm->trapframe = (struct trapframe *)kalloc()) == 0){
  //   ER();
  // }

  // 申请内核栈
  // if((mm->kstack = (uint64_t)kalloc()) == 0){
  //   ER();
  // }
  // printf(rd("kstack: %p\n"), mm->kstack);

  /* 创建页表并映射内核空间, trapframe, kstack */
  // sys_memuse();
  if((mm->pagetable = proc_pagetable(mm, tsk)) == NULL)
    ER();
  // sys_memuse();

  return mm;
}

static void init_new_context(proc_t *tsk){
  // Set up new context to start executing at forkret,
  // which returns to user space.
  memset(&tsk->context, 0, sizeof(tsk->context));
  tsk->context.ra = (uint64_t)forkret;
  /* 通过虚拟地址或者物理地址来映射内核栈 */
  tsk->context.sp = tsk->kstack + PGSIZE;
  // tsk->context.sp = KSTACK + PGSIZE;
}

/**
 * @brief 给tsk分配新的mm，把当前进程的mm复制给它，
 * 如果为lightweight进程，则和父进程共用一个mm.
 * 
 */
static int copy_mm(uint64_t clone_flags, proc_t * tsk){
  mm_struct_t *mm, *oldmm;
  proc_t *current = myproc();

  oldmm = current->mm;
  if(!oldmm)
    ER(); 

  /* 轻量级进程，亲子共用一个mm_struct*/
  if(clone_flags & CLONE_VM){
    ERROR("not implemented!");
  }

  if((mm = kzalloc(sizeof(mm_struct_t))) == NULL)
    ER();

  memcpy(mm, oldmm, sizeof(*mm));

  /* 给mm分配页表，并且映射内核空间 */
  // sys_memuse();
  if(!mm_init(mm, tsk))
    ER();

  /* 复制vma和用户空间的页表 */
  // sys_memuse();
  dup_mmap(mm, oldmm);
  // sys_memuse();
  // 复制trapframe
  // *(mm->trapframe) = *(oldmm->trapframe);

  // 将返回地址重置为0
  // mm->trapframe->a0 = 0;

  tsk->mm = mm;
  return 0;
}

// Create a new process, copying the parent.
// Sets up child kernel stack to return as if from fork() system call.
int
do_clone(uint64_t stack)
{
  // sys_memuse();
  int i, pid;
  struct proc *np;
  struct proc *p = myproc();

  // Allocate process.
  if((np = allocproc()) == 0){
    return -1;
  }

  // sys_memuse();
  copy_mm(0, np);
  // sys_memuse();

  init_new_context(np);

  // 拷贝内存布局(如果开启了COW，那么仅仅是复制页表)
  // if(uvmcopy(p->pagetable, np->pagetable, p->sz) < 0){
  //   freeproc(np);
  //   release(&np->lock);
  //   return -1;
  // }
  // np->sz = p->sz;

  // 复制trapframe
  *(np->trapframe) = *(p->trapframe);

  // 将返回地址重置为0
  np->trapframe->a0 = 0;

  // 如果指定了栈，那么重设sp
  if(stack)
    np->trapframe->sp = stack;

  // 增加文件描述符引用
  for(i = 0; i < NOFILE; i++)
    if(p->ofile[i])
      np->ofile[i] = filedup(p->ofile[i]);
  np->cwd = edup(p->cwd);

  safestrcpy(np->name, p->name, sizeof(p->name));

  pid = np->pid;

  release(&np->lock);

  acquire(&wait_lock);
  np->parent = p;
  release(&wait_lock);

  acquire(&np->lock);
  np->state = RUNNABLE;
  release(&np->lock);


  // np->cur_mmap_sz = p->cur_mmap_sz;
  // dup_mmap(np->mm, p->mm);

  // vmprint(p->pagetable);
  // vmprint(np->pagetable);

  // sys_memuse();
  return pid;
}

// Pass p's abandoned children to init.
// Caller must hold wait_lock.
void
reparent(struct proc *p)
{
  struct proc *pp;

  for(pp = proc; pp < &proc[NPROC]; pp++){
    if(pp->parent == p){
      pp->parent = initproc;
      wakeup(initproc);
    }
  }
}


/**
 * @brief free the mm_struct, in linux it's no be freeed in this func
 * 
 * @param mm 
 */
static void
mmput(mm_struct_t *mm){
  // if(mm->trapframe)
  //   kfree((void*)mm->trapframe);
  // if(mm->kstack)
  //   kfree((void*)mm->kstack);
  // mm->trapframe = 0;
  // mm->kstack = 0;

  /* unmap内核空间 */
  proc_freepagetable(mm->pagetable);
  /* unmap 用户空间并释放对应的vma */
  exit_mmap(mm);

  // vmprint(mm->pagetable);
  /* 重要！之前是在uvmfree中调用的 */
  freewalk(mm->pagetable);

  kfree(mm);
}

/**
 * @brief 在free_proc函数中调用
 * 
 */
void 
exit_mm(struct proc *tsk){
  mm_struct_t *mm = tsk->mm;

  if(!mm)
    return;
  
  tsk->mm = NULL;
  mmput(mm);
}

// Exit the current process.  Does not return.
// An exited process remains in the zombie state
// until its parent calls wait().
void
exit(int status)
{
  struct proc *p = myproc();

  if(p == initproc) {
    panic("init exiting");
  }

  if(status == -1) {
    panic("exception quit");
  }
  // Close all open files.
  for(int fd = 0; fd < NOFILE; fd++){
    if(p->ofile[fd]){
      struct file *f = p->ofile[fd];
      fileclose(f);
      p->ofile[fd] = 0;
    }
  }

  eput(p->cwd);
  p->cwd = 0;

  /* free the mm field of the process 
  释放页表和结构体，在waitpid中释放*/
  // exit_mm(p);

  acquire(&wait_lock);

  // Give any children to init.
  reparent(p);

  // Parent might be sleeping in wait().
  wakeup(p->parent);
  
  acquire(&p->lock);

  p->xstate = status;
  p->state = ZOMBIE;

  release(&wait_lock);

  // Jump into the scheduler, never to return.
  sched();
  panic("zombie exit");
}

// Wait for a child process to exit and return its pid.
// Return -1 if this process has no children.
int
waitpid(int cid, uint64 addr)
{
  struct proc *np;
  int havekids, pid, haveckid;
  struct proc *p = myproc();

  acquire(&wait_lock);

  for(;;){
    // Scan through table looking for exited children.
    havekids = 0;
    haveckid = 0;
    for(np = proc; np < &proc[NPROC]; np++){
      if(np->parent == p){
        // make sure the child isn't still in exit() or swtch().
        acquire(&np->lock);

        havekids = 1;

        pid = np->pid;
        if(cid != -1 && cid != pid) continue;

        haveckid = 1;
        if(np->state == ZOMBIE){
          // Found one.
          
          if(addr != 0 && copyout(p->mm->pagetable, addr, (char *)&np->xstate,
                                  sizeof(np->xstate)) < 0) {
            release(&np->lock);
            release(&wait_lock);
            return -1;
          }
          freeproc(np);
          release(&np->lock);
          release(&wait_lock);
          return pid;
        }
        release(&np->lock);
      }
    }

    // No point waiting if we don't have any children.
    if(!havekids || !haveckid || p->killed){
      release(&wait_lock);
      return -1;
    }
    
    // Wait for a child to exit.
    sleep(p, &wait_lock);  //DOC: wait-sleep
  }
}

// Per-CPU process scheduler.
// Each CPU calls scheduler() after setting itself up.
// Scheduler never returns.  It loops, doing:
//  - choose a process to run.
//  - swtch to start running that process.
//  - eventually that process transfers control
//    via swtch back to the scheduler.
void
scheduler(void)
{
  struct proc *p;
  struct cpu *c = mycpu();
  
  c->proc = 0;
  for(;;){
    // Avoid deadlock by ensuring that devices can interrupt.
    intr_on();

    for(p = proc; p < &proc[NPROC]; p++) {
      acquire(&p->lock);
      if(p->state == RUNNABLE) {
        // Switch to chosen process.  It is the process's job
        // to release its lock and then reacquire it
        // before jumping back to us.
        p->state = RUNNING;
        c->proc = p;
        switchuvm(p->mm);
        swtch(&c->context, &p->context);
        switchkvm();
        // Process is done running for now.
        // It should have changed its p->state before coming back.
        c->proc = 0;
      }
      release(&p->lock);
    }
  }
}

// Switch to scheduler.  Must hold only p->lock
// and have changed proc->state. Saves and restores
// intena because intena is a property of this
// kernel thread, not this CPU. It should
// be proc->intena and proc->noff, but that would
// break in the few places where a lock is held but
// there's no process.
void
sched(void)
{
  int intena;
  struct proc *p = myproc();

  if(!holding(&p->lock))
    panic("sched p->lock");
  if(mycpu()->noff != 1)
    panic("sched locks");
  if(p->state == RUNNING)
    panic("sched running");
  if(intr_get())
    panic("sched interruptible");

  intena = mycpu()->intena;
  swtch(&p->context, &mycpu()->context);
  mycpu()->intena = intena;
}

// Give up the CPU for one scheduling round.
void
yield(void)
{
  struct proc *p = myproc();
  acquire(&p->lock);
  p->state = RUNNABLE;
  sched();
  release(&p->lock);
}

#include "fs/fat.h"
// A fork child's very first scheduling by scheduler()
// will swtch to forkret.
void
forkret(void)
{
  static int first = 1;

  // Still holding p->lock from scheduler.
  release(&myproc()->lock);

  if (first) {
    // File system initialization must be run in the context of a
    // regular process (e.g., because it calls sleep), and thus cannot
    // be run from main().
    first = 0;

    extern fat32_t *fat;
    printf("ready to mount\n");
    fat_mount(ROOTDEV, &fat);
    printf("mount done\n");
    myproc()->cwd = namee(NULL, "/");
    printf("ok\n");
    // LOOP();
    // fsinit(ROOTDEV);


  }

  usertrapret();
}

// Atomically release lock and sleep on chan.
// Reacquires lock when awakened.
void
sleep(void *chan, struct spinlock *lk)
{
  struct proc *p = myproc();
  
  // Must acquire p->lock in order to
  // change p->state and then call sched.
  // Once we hold p->lock, we can be
  // guaranteed that we won't miss any wakeup
  // (wakeup locks p->lock),
  // so it's okay to release lk.

  if(lk != &p->lock){
    acquire(&p->lock);
    release(lk);
  }

  // Go to sleep.
  p->chan = chan;
  p->state = SLEEPING;

  sched();

  // Tidy up.
  p->chan = 0;

  // Reacquire original lock.
  release(&p->lock);
  acquire(lk);
}

// Wake up all processes sleeping on chan.
// Must be called without any p->lock.
void
wakeup(void *chan)
{
  struct proc *p;

  for(p = proc; p < &proc[NPROC]; p++) {
    if(p != myproc()){
      acquire(&p->lock);
      if(p->state == SLEEPING && p->chan == chan) {
        p->state = RUNNABLE;
      }
      release(&p->lock);
    }
  }
}

// Kill the process with the given pid.
// The victim won't exit until it tries to return
// to user space (see usertrap() in trap.c).
int
kill(int pid)
{
  struct proc *p;

  for(p = proc; p < &proc[NPROC]; p++){
    acquire(&p->lock);
    if(p->pid == pid){
      p->killed = 1;
      if(p->state == SLEEPING){
        // Wake process from sleep().
        p->state = RUNNABLE;
      }
      release(&p->lock);
      return 0;
    }
    release(&p->lock);
  }
  return -1;
}

// Copy to either a user address, or kernel address,
// depending on usr_dst.
// Returns 0 on success, -1 on error.
int
either_copyout(int user_dst, uint64 dst, void *src, uint64 len)
{
  struct proc *p = myproc();
  if(user_dst){
    return copyout(p->mm->pagetable, dst, src, len);
  } else {
    memmove((char *)dst, src, len);
    return 0;
  }
}

// Copy from either a user address, or kernel address,
// depending on usr_src.
// Returns 0 on success, -1 on error.
int
either_copyin(void *dst, int user_src, uint64 src, uint64 len)
{
  struct proc *p = myproc();
  if(user_src){
    return copyin(p->mm->pagetable, dst, src, len);
  } else {
    memmove(dst, (char*)src, len);
    return 0;
  }
}

// Print a process listing to console.  For debugging.
// Runs when user types ^P on console.
// No lock to avoid wedging a stuck machine further.
void
procdump(void)
{
  static char *states[] = {
  [UNUSED]    "unused",
  [SLEEPING]  "sleep ",
  [RUNNABLE]  "runble",
  [RUNNING]   "run   ",
  [ZOMBIE]    "zombie"
  };
  struct proc *p;
  char *state;

  printf("\n");
  for(p = proc; p < &proc[NPROC]; p++){
    if(p->state == UNUSED)
      continue;
    if(p->state >= 0 && p->state < NELEM(states) && states[p->state])
      state = states[p->state];
    else
      state = "???";
    printf("%d %s %s", p->pid, state, p->name);
    printf("\n");
  }
}
